/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008, 2012. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.rqm.adapter.commandline;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.concurrent.CountDownLatch;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.ibm.rqm.adapter.library.connector.AbstractTestToolAdapter;
import com.ibm.rqm.adapter.library.connector.IQMToolAdapter;
import com.ibm.rqm.adapter.library.data.AbstractLogEvent;
import com.ibm.rqm.adapter.library.data.AdapterConstants;
import com.ibm.rqm.adapter.library.data.CustomProperty;
import com.ibm.rqm.adapter.library.data.ElementTypeEnum;
import com.ibm.rqm.adapter.library.data.ILogEvent;
import com.ibm.rqm.adapter.library.data.ILogResult;
import com.ibm.rqm.adapter.library.data.ResultStateEnum;
import com.ibm.rqm.integration.client.clientlib.Util;


/**
 * The entity which will execute the tasks which has been taken up by the
 * adapter.<br> 
 * The tasks that the adapter can perform are registered with the
 * adapter web plug-in registry as capabilities <br>( extension point
 * <code>com.ibm.rqm.planning.common.executionScriptType</code><br>
 * property <code>capability</code>)
 * <br>
 * eg for plug-in registry.<br> 
 * <code>
 * &lt;extension
 *     point="com.ibm.rqm.planning.common.executionScriptType"&gt;
 *  		&lt;type
 *          adapterName="Sample: Command Line Adapter"
 *          description="RQM-KEY-CMD-APTR-TYPE-DESC"
 *          id="com.ibm.rqm.adapter.commandline"
 *          name="RQM-KEY-CMD-APTR-TYPE-NAME"&gt;
 *       &lt;capability
 *             name="EXECUTE"&gt;
 *       &lt;/capability&gt;
 *       &lt;capability
 *             name="IMPORT"&gt;
 *       &lt;/capability&gt;
 *   		&lt;/type&gt;
 *	&lt;/extension&gt;
 * </code>
 * <br>
 * For a capability <code>IMPORT</code> the method to implement is {@link IQMToolAdapter#getResources(String)}
 * For a capability <code>EXECUTE</code> the method to implement is {@link IQMToolAdapter#runTest(String[])}
 * 
 * Note: At present (in 1.0) these are the only two capabilities supported by the framework
 * 
 * @author amathur
 */
public class CommandLineAdapter extends AbstractTestToolAdapter {

	private File stdOutFile = null;
	private File stdErrFile = null;
	private File execVariablesFile = null;
	private File custPropertiesFile = null;
	private File execPropertiesFile = null;
	private String errorMessage = "";
	private Process process = null;
	private TrackProgress trackProgress= null;
	
	private LogProgress logProgress = null;
	
	public static final int STATUS_NOTSTARTED = 0;
	public static final int STATUS_INITIALIZED = 1;
	public static final int STATUS_EXECUTING = 2;
	public static final int STATUS_EXECUTION_DONE = 3;
	public static final int STATUS_DONE = 4;
	public static final int STATUS_STOPPED = 5;
	
	int status = STATUS_NOTSTARTED;
	
	private static final String CREATE_EXEC_PROPERTIES = "CREATE";
	private final CountDownLatch execPropertiesAreSet = new CountDownLatch(1);
	
	public static boolean winOS = isWinOs();
	public static final HashSet<String> commands = CommandLineAdapter.initAllowedExecutablesSet();
	
	/**
	 * Import is not supported in the Command line adapter. All extensions
	 * should override this method to provide support for import.
	 * 
	 * @param path -
	 *            The input location on the adapter machine
	 * @return - an array of strings with the name of resources.
	 */
	public String[] getResources(String path) {
		return null;
	}
	
	public static class MyFd extends Thread {
		private InputStream fd;
		private File f;
		public MyFd(InputStream fd,File f) {
			this.fd=fd;
			this.f=f;
		}
		public void run() {
			byte[] bytes = new byte[4096];
			try {
				BufferedInputStream bis = new BufferedInputStream(fd);
				BufferedOutputStream out = new BufferedOutputStream(new FileOutputStream(f));
				int amount=0;
				while ((amount=bis.read(bytes)) >= 0) {
		        	out.write(bytes,0,amount);
		        }
				bis.close();
				out.close();
				fd.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}
	
	public void setErrorMessage(String mesg)
	{
		errorMessage = mesg;
	}
	
	public String getErrorMessage()
	{
		return errorMessage;
	}
	
	/* (non-Javadoc)
	 * @see com.ibm.rqm.adapter.library.connector.IQMToolAdapter#runTest(java.lang.String[])
	 */
	public void runTest(String[] args)  {
		runTest(args, null, false);
	}
	
	public void runTest(String[] args, Properties executionVariables, boolean runningJUnit) {
		setStatus(STATUS_NOTSTARTED);
		StringBuffer commandLine = new StringBuffer();
		if ('\\' == File.separatorChar){
			commandLine.append("\"");
			commandLine.append(args[0]);
			commandLine.append("\"");
		} else {
			commandLine.append(args[0]);
		}
		if (args[1] != null) {
			commandLine.append(" ");
			commandLine.append(args[1]);
		}
		
		try {
			
			if (!runningJUnit) 
				// TODO Doing this opens security hole.  Need to design way to allow restricting java apps and classpath
				validateExe(new File(args[0]).getName());
			
			List<AbstractLogEvent> logEvents = new ArrayList<AbstractLogEvent>();
			
			trackProgress = new TrackProgress();
			AbstractLogEvent event = new CLALogEvent();
			setStatus(STATUS_INITIALIZED);
			ResultStateEnum result = ResultStateEnum.ERROR;
			try {
				trackProgress.start();
				createOutFiles();
				event.starttime = System.currentTimeMillis();
				// Start command line execution and track progress ...
				process = startProcess(commandLine.toString(), executionVariables);
				trackProgress.setProcess(process);
				MyFd outFd=new MyFd(process.getInputStream(),stdOutFile);outFd.start();
				MyFd errFd=new MyFd(process.getErrorStream(),stdErrFile);errFd.start();
				setStatus(STATUS_EXECUTING);
				int exitValue = process.waitFor();
				setStatus(STATUS_EXECUTION_DONE);
				outFd.join(3000);
				errFd.join(3000);
				trackProgress.stop();
				System.out.println("ExitCode: "+exitValue); // show result code in console!!!!
				result = CommandLinePreferences.prefs().getExitCode(exitValue);
				((CommandLineLogResult)logResult).setStatus(result!= null ? result : ResultStateEnum.UNKNOWN);
				// Use notes to add link to any external data or testers note about execution
//				((CommandLineLogResult)logResult).setNotes("<a href=\"http://google.com\">google</a><br/><br/><br/>");
				// Load additional system properties as needed
				((CommandLineLogResult)logResult).setCustomProperties(getCustomProperties());
				((CommandLineLogResult)logResult).setExecutionVariables(executionVariables);
			} catch (Exception e) {
				if (this.status == STATUS_STOPPED)
					throw e;
				Logger.Log.error(e.getMessage(), e);		
				((CommandLineLogResult)logResult).setStatus(result);
				((CommandLineLogResult)logResult).setNotes(e.getMessage());
				MyFd errFd=new MyFd(new ByteArrayInputStream((e.getMessage()+getErrorMessage()).getBytes("UTF-8")),stdErrFile);errFd.start();
				errFd.join();
			} finally {
				trackProgress.stop();
			}
			event.resultState = result;
			event.type = ElementTypeEnum.REPORTING;
			event.endtime = System.currentTimeMillis();
			logEvents.add(event);

			((CommandLineLogResult)logResult).setLogEvents(logEvents.toArray(new ILogEvent[] {}));
		} catch (Exception e) {
			Logger.Log.error(e.getMessage(), e);	
			((CommandLineLogResult)logResult).setNotes(e.getMessage());
			((CommandLineLogResult)logResult).setStatus(ResultStateEnum.ERROR);
		}
	}

	public void createOutFiles() throws IOException {
		stdOutFile=File.createTempFile("QMAdapter",".out");  //$NON-NLS-1$//$NON-NLS-2$
		stdErrFile=File.createTempFile("QMAdapter",".err");  //$NON-NLS-1$//$NON-NLS-2$
	}
	
	public void cleanupOutFiles() {
		if(stdOutFile!=null) {
			stdOutFile.delete();
		}
		if(stdErrFile!=null) {
			stdErrFile.delete();
		}
	}
	
	public void setLogResult(ILogResult logResult){
		this.logResult = logResult;
	}

	public void setLogProgress(LogProgress logProgress){
		this.logProgress = logProgress;
	}

	public File getStdOutFile() {
		return stdOutFile;
	}

	public File getStdErrFile() {
		return stdErrFile;
	}
	
	public boolean uploadError(String errorMessage, ILogResult logResult){
		ByteArrayInputStream bais = null;
		byte[] messageBytes;
		try{
			try {
				messageBytes = errorMessage.toString().getBytes("UTF-8");
			} catch (UnsupportedEncodingException e1) {
				messageBytes = errorMessage.toString().getBytes();
			}
			bais = new ByteArrayInputStream(messageBytes);
			File stdErrFile = getStdErrFile();
			if (stdErrFile == null){
				createOutFiles();
				stdErrFile = getStdErrFile();
			}
			CommandLineAdapter.MyFd stdErrorThread = new CommandLineAdapter.MyFd(bais, stdErrFile);stdErrorThread.start();
			stdErrorThread.join();
			try{
				if (stdErrFile != null && stdErrFile.getCanonicalFile().length() != 0){
					logResult.addAttachment(stdErrFile.getCanonicalPath(), "text/plain"); //$NON-NLS-1$
				}
			}
			catch(IOException e){
				Logger.Log.error(e.getMessage(),e);
			}
			return true;
		} catch (Exception e3){
			return false;
		} 
	}
	
	/*
	 * Setup environment variables, Properties files, and attachments file.
	 */
	private Process startProcess(String commandline, Properties vars) throws IOException {
		ArrayList<String> cmd = new ArrayList<String>();
		/*
		 * Use Pattern/Matcher so quoted args will be stored in a single element in the cmd list.
		 * We strip off the quotes so they are not passed to the executing command (this was the 
		 * cause of the problem seen in WI66114).  
		 */
		Pattern p = Pattern.compile("\"[^\"]+\"|\\S+"); //$NON-NLS
		Matcher m = p.matcher(commandline);
		while (m.find()) {
			// Strip off any leading/trailing quotes and add to cmd
			cmd.add(Util.strip(m.group(), "\"'")); //$NON-NLS
		}

		ProcessBuilder pb = new ProcessBuilder(cmd);
		Map<String, String> env = pb.environment();
		
			
		// Create file that will be used to pass execution variable values back from the invoked command.  
		setExecVariablesFile(File.createTempFile("QMAdapter",".vars"));  //$NON-NLS-1$//$NON-NLS-2$
		String execVarsPath = getExecVariablesFile().getPath();
		// Add environment variable specifying the name of the file
		env.put("qm_ExecutionVariablesFile", execVarsPath);
		if ( vars != null ) {
			OutputStream varsOut = new FileOutputStream(execVarsPath);
			vars.store(varsOut, "Execution Variables"); //$NON-NLS
			varsOut.close();
			
			// Create an environment variable for each execution variable specified
			Enumeration<Object> keys = vars.keys();
			while (keys.hasMoreElements()) {
				String key = (String)keys.nextElement();
				String value = (String)vars.getProperty(key);
				env.put("qm_"+cleanPropertyName(key), value); //$NON-NLS-1$
			}
		}

		// Create file that will be used to pass custom properties back from the invoked command.
		setCustPropertiesFile(File.createTempFile("QMAdapter", ".props"));    //$NON-NLS-1$//$NON-NLS-2$
		String custPropertiesPath = getCustPropertiesFile().getPath();
		env.put("qm_CustomPropertiesFile",custPropertiesPath);
		
		// Set the path to the execution properties file if specified.
		File execPropFile = getExecPropertiesFile();
		if (execPropFile != null)
			env.put("qm_ExecutionPropertiesFile", execPropFile.getPath());
		
		/*
		 * Create file that will be used to specify files/links to attach to the
		 * execution result. Check if user specified a path via
		 * RESULT_ATTACHMENT_FILE_PROPERTY. If not, create temp file and set
		 * property to point there.
		 */
		String attachmentsPath = CommandLinePreferences.prefs().getProperty(AdapterConstants.RESULT_ATTACHMENT_FILE_PROPERTY, "");
		File f = null;
		if (attachmentsPath != null && attachmentsPath.length() > 0) {
			f = new File(attachmentsPath);
			if (f.exists()) {
				if (f.isDirectory()) {
					f = File.createTempFile("QMAdapter", ".attach"); //$NON-NLS-1$//$NON-NLS-2$
					// TODO Externalize this message
					Logger.Log.error(attachmentsPath + " is a directory, using attachments file " + f.getCanonicalPath());
				}
				else {
					f.delete();
					f.createNewFile();
				}
			}
			else {
				f.createNewFile();
			}
		}
		else {
			f = File.createTempFile("QMAdapter", ".attach"); //$NON-NLS-1$//$NON-NLS-2$
		}
		attachmentsPath = f.getCanonicalPath();
		// TODO: Externalize this string?
		String comment = "# Be sure to escape any backslash characters in pathnames.  For example, MyAttachedFile=c:\\\\tmp\\\\log.txt\n"; //$NON-NLS-1$
		FileOutputStream fos = new FileOutputStream(attachmentsPath);
		fos.write(comment.getBytes("UTF-8"));
		fos.close();
		env.put("qm_AttachmentsFile", attachmentsPath); //$NON-NLS-1$
		System.setProperty(AdapterConstants.RESULT_ATTACHMENT_FILE_PROPERTY, attachmentsPath);
		
		return pb.start();
	}
	
	private String cleanPropertyName(String name) {
		char[] chars = name.toCharArray();
		int length = (chars != null ? chars.length : 0);
		for (int i = 0; i < length; ++i) {
			if (!Character.isJavaIdentifierPart(chars[i]))
				chars[i] = '_';
		}
		return new String(chars);
	}
	
	/*
	 * Get custom properties provided by the user in the preferences file and optionally add in
	 * any system properties that are requested.
	 */
	private CustomProperty[] getCustomProperties() {
		Map<String,String> props = CommandLinePreferences.prefs().getNonRqmProperties();
		if (CommandLinePreferences.prefs().getBooleanProperty(CommandLinePreferences.INCLUDE_SYSTEM_PROPERTIES, false)) {
			Properties systemProps = System.getProperties();
			Enumeration<Object> keys = systemProps.keys();
			while (keys.hasMoreElements()) {
				String key = (String)keys.nextElement();
				if ( key.startsWith("os.") || key.startsWith("user.") )
					props.put(key, systemProps.getProperty(key));
			}
		}
		
		//grab custom properties (if there are any from command execution and append to props.
		try {
			FileInputStream scriptCustPropertyFile = new FileInputStream(getCustPropertiesFile());
			if (scriptCustPropertyFile != null) {
				Properties scriptProperties = new java.util.Properties();
				scriptProperties.load(scriptCustPropertyFile);
				// file no longer needed now
				scriptCustPropertyFile.close();
				getCustPropertiesFile().delete();
				Enumeration<Object> keys = scriptProperties.keys();
				while (keys.hasMoreElements()) {
					String key = (String)keys.nextElement();
					props.put(key, scriptProperties.getProperty(key));
				}
			}
		}
		catch(IOException e) {  // no custom properties file found
		}
		
		CustomProperty[] result = new CustomProperty[props.size()];
		Set<String> keys = props.keySet();
		Iterator<String> iter = keys.iterator();
		int index = 0;
		while (iter.hasNext()) {
			String key = iter.next();
			CustomProperty prop = new CustomProperty();
			prop.setPropertyName(key);
			String value = props.get(key);
			value.replaceAll("<", "&lt;");
			value.replaceAll(">", "&gt;");
			prop.setPropertyValue(value);
			//System.out.println(key +" : "+value);
			prop.setNameSpaceWithPrefix("rqmprop", "http://jazz.net/xmlns/alm/rqm/common");
			result[index++] = prop;
		}
		return result;
	}
	
	public void stopExecution(){
		
		
		setStatus(STATUS_STOPPED);
		
		if(process!= null)
			process.destroy();
		
    	if(this.trackProgress!=null)
 		   trackProgress.stop();	
    	
    	
		
	}
	
	public synchronized int setStatus(int setStatus)
	{
		if (this.status == STATUS_EXECUTION_DONE)
			return this.status;
		
		if (setStatus == STATUS_STOPPED)
		{
			this.status = setStatus;
			return status;
		}
		
		if (this.status == STATUS_STOPPED)
			throw new ExecutionStoppedException();
		
		this.status = setStatus;
		return setStatus;
			
	}
	
	private class TrackProgress {
		
		Thread thread = null;
		boolean running = false;
		Process process = null;
		
		public synchronized Thread start() {
			if (!running) {
				running = true;
				thread = new Thread(new Runnable() {
					public void run() {
						monitorExecution();
						running = false;
					}
				});
				thread.setDaemon(true);
				thread.setName("CommandLineAdapter.TrackProgress");
				thread.start();
			}
			return thread;
		}
		
		public synchronized void stop() {
			if ( running ) {
				running = false;
				thread.interrupt();
			}
		}
		
		public void setProcess(Process process) {
			this.process = process;
		}
		
		private void monitorExecution() {
			String monitorProperties = CommandLinePreferences.prefs().getProperty(CommandLinePreferences.EXECUTION_PROPERTIES, "");
			File monitor = null;
			if (monitorProperties != null && monitorProperties.length() > 0) {
				// Check for special value indicating we should create the execution properties file.  Used by
				// Selenium adapter which handles progress updates itself.
				try {
					if (CREATE_EXEC_PROPERTIES.equals(monitorProperties)) {
						monitor = File.createTempFile("QMAdapter", ".execprop");
					}
					else {
						monitor = new File(monitorProperties);
						// If execution properties file already exists, clear it and start fresh
						if (monitor.exists()) {
							monitor.delete();
						}
						monitor.createNewFile();
					}
				} catch (IOException e) {
					System.out.println("Monitor thread failed to create monitor file: [" + monitor.getPath() + "], " + e.getLocalizedMessage());
					monitor = null;
				}
			}
			setExecPropertiesFile(monitor);
			if ( logProgress != null && monitor != null) {
				System.out.println("Monitor Progress: ["+monitor.getPath()+"]");
				long since = -1;
				while (running) {
					try {
						// start monitoring after first polling interval completes ....
						Thread.sleep(5000); // use normal ping interval?					

						//process.
						if ( process == null )
							continue; // process has not been started yet ...
						
						// Check if there is a status file being written by the application
						// under test and if so if it has changed.  If changed then load and
						// update the current execution status.
						if ( monitor.exists() && since != monitor.lastModified() ) {
							Properties props = new Properties();
							try {
								FileInputStream stream = new FileInputStream(monitor);
								props.load(stream);
								stream.close();
								String comment = "<progress must be in 0..99>";
								int progress = parseInt(props.getProperty("progress", "0"));
								if ( progress >= 0 && progress <= 99 ) {
									comment = props.getProperty("comment", "");
									logProgress.logProgress(progress, comment, null);
								}
								System.out.println("Progress: "+progress+": "+comment);
							} catch (IOException ioexc) {
								System.out.println(ioexc.toString());
							}
							since = monitor.lastModified();
						}
					} catch (InterruptedException e) {
						if (running)
							System.out.println("Monitor thread interrupted");
					}
				}
			}
		}
		
		private boolean isProcessAlive(Process process) {
			if ( process == null ) return true;  // too soon, not registered yet ...
			boolean result = true;
			try {
				process.exitValue(); // throws IllegalThreadStateException if halted ...
			} catch ( IllegalThreadStateException itse ) {
				result = false;
			}
			return result;
		}
		
		private int parseInt(String number) {
			int result = 0;
			try {
				result = (number != null ? Integer.parseInt(number.trim()) : 0);
			} catch (NumberFormatException numFormat) {
			}
			return result;
		}
		
	}
	public static HashSet<String>  initAllowedExecutablesSet()	{
		String allowedExecutables = CommandLinePreferences.prefs().getProperty(CommandLinePreferences.ALLOWED_EXECUTABLES, "");
		if(allowedExecutables == null || allowedExecutables.length()==0)
			return null;
		StringTokenizer st = new StringTokenizer(allowedExecutables, ",");
		HashSet<String> hs = new HashSet<String>();
		while(st.hasMoreTokens()){
			 if(winOS)
			   hs.add(st.nextToken().trim().toLowerCase()); //for win OS executables case insensitive
			 else
			    hs.add(st.nextToken().trim());
		}
		return hs;
	}
	
	public static boolean isWinOs(){
		if('/' == File.separatorChar)
			return false;
		else
			return true;		
	}
	
	private void validateExe(String command) throws Exception{		
		
		if(winOS) //for win OS executables case insensitive
			command = command.toLowerCase(); 
		
		if( commands!= null && !commands.contains(command)) //need to validate against HashSet, if not set any command allowed		
			throw new Exception( command + " is not an allowed program specified in " + QMCommandLineAdapter.DEFAULT_PREFERENCES_NAME);				
	}

	public File getExecVariablesFile() {
		return execVariablesFile;
	}

	public void setExecVariablesFile(File execVariablesFile) {
		this.execVariablesFile = execVariablesFile;
	}
	
	public File getCustPropertiesFile()	{
		return custPropertiesFile;
	}
	
	public void setCustPropertiesFile(File custPropertiesFile) {
		this.custPropertiesFile = custPropertiesFile;
	}

	public File getExecPropertiesFile() {
		try {
			this.execPropertiesAreSet.await();
		} catch (InterruptedException e) {
			// If we get interrupted, something is probably wrong, just return current value, probably null
		}
		return execPropertiesFile;
	}

	public void setExecPropertiesFile(File execPropertiesFile) {
		this.execPropertiesFile = execPropertiesFile;
		this.execPropertiesAreSet.countDown();
	}
}
