/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008, 2012. All Rights Reserved.
 *  
 * U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 ******************************************************************************/
package com.ibm.rqm.adapter.commandline;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.net.URLEncoder;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.apache.abdera.model.Element;

import com.ibm.rqm.adapter.library.data.AdapterConstants;
import com.ibm.rqm.adapter.library.data.AdapterException;
import com.ibm.rqm.adapter.library.data.NewRequester;
import com.ibm.rqm.adapter.library.data.ResultStateEnum;

public class CommandLineExecutionThread extends Thread{

	
	private CommandLineAdapter toolAdapter;
	private CommandLineLogResult logResult;
	private String url;
	private NewRequester requester;
	private Element element;
	private QMCommandLineProcessor processor;
	private String taskId= "";
	
	private final String RESOURCE_URN = "resourceURN=";


	public CommandLineExecutionThread(String url, Element element, NewRequester requester, QMCommandLineProcessor processor){
		
		this.url = url;
		this.element= element;
		this.requester = requester;
		this.processor = processor;
		taskId = requester.getValue(AdapterConstants.DC_PREFIX+":identifier", element);
	}
	
	public  void run() {
		synchronized(requester){		
		try{
			processor.setBusy(true);
			toolAdapter = new CommandLineAdapter();
			logResult = new CommandLineLogResult();
			((CommandLineAdapter)toolAdapter).setLogResult(logResult);
			String command = null;
			String relativepath = null;
			String fullpath = null;
			File tempFile = null;
			File parentDirectory = null;
			boolean success = true;
			boolean isJUnitSelenium = requester.getAdapter().getType().equals(QMCommandLineAdapter.SELENIUM_ADAPTER_TYPE);
			File junitReportFile = null;
			File junitScreenshotDir = null;
			LogProgress progress = new LogProgress(url, requester);
			((CommandLineAdapter)toolAdapter).setLogProgress(progress);
			boolean executionMonitor =  false;
			String monitorProperties = CommandLinePreferences.prefs().getProperty(CommandLinePreferences.EXECUTION_PROPERTIES, "");
			if(monitorProperties != null && !monitorProperties.equals(""))
				executionMonitor = true;
			try {
				progress.logProgress(0, "Starting ...", null);
				String args = requester.getValue("qmtask:arguments", element);

				// Is this is local resource or a shared?
				String manage = requester.getValue("qmtask:managedbyadapter", element);
				command = requester.getValue("qmtask:command", element);
				fullpath = requester.getValue("qmtask:fullpath", element);
				relativepath = requester.getValue("qmtask:relativepath", element);

				if ((command == null) || ("".equals(command)) || ("null".equals(command))) {
					// Use the relative path instead because command is null
					command = null;
				}

				if ((fullpath == null) || ("".equals(fullpath)) || ("null".equals(fullpath))) {
					// Use the fullpath instead because command is null
					fullpath = null;
				}

				if ((relativepath == null) || ("".equals(relativepath)) || ("null".equals(relativepath))) {
					// Use the relative path instead because command is null
					relativepath = null;
				}

				// there are a couple of cases where 'command' may be not set
				// a script created by importing from an adapter
				// a non-browsed, typed-in shared resource

				if ((manage != null) && ("false".equals(manage))) {
					// Shared Resource
					// Create a local temp name for the resource
					String tempName = command;
					if (tempName == null) {
						tempName = relativepath;
					}
					if (tempName == null) {
						tempName = fullpath;
					}

					// Now get the file name from the path
					String[] parts = tempName.split("[\\\\/:]");
					if (parts.length > 0) {
						tempName = parts[parts.length - 1];
					}

					// Get the resource
					String resourceUrl = requester.getLinkHref("qm:resource", element);

					// Verify the URL is good 
					// If not, use the command
					if (resourceUrl != null) {
						String urlParts[] = resourceUrl.split("=");
						if ((urlParts.length == 1) && (resourceUrl.endsWith("="))) {
							resourceUrl = resourceUrl.concat(command);
						}
					}
					FileOutputStream os = null;
					// zero out the command since it needs to be the local temp file
					command = null;

					int indexOfUrn = resourceUrl.indexOf(RESOURCE_URN);
					String resource = resourceUrl;
					if (indexOfUrn > -1 && resourceUrl.length() > indexOfUrn + RESOURCE_URN.length()){
						resource = resourceUrl.substring(indexOfUrn + RESOURCE_URN.length());
					}
					progress.logProgress(5, "Downloading resource ...", resource);
					// Now create the local file
					try {
						// the file  will be downloaded to a directory named as adapter id inside the user temp folder
						StringBuffer sb = new StringBuffer(System.getProperty("java.io.tmpdir"));
						if (!sb.toString().endsWith(File.separator)){
							sb.append(File.separator);
						}
						sb.append(System.currentTimeMillis());
						String parentPath = sb.toString();
						parentDirectory = new File(parentPath);
						if (!parentDirectory.exists()){
							parentDirectory.deleteOnExit();
							parentDirectory.mkdirs();
						}
						if (!parentDirectory.isDirectory()){
							parentPath = parentPath + System.currentTimeMillis();
							parentDirectory = new File(parentPath); 
							parentDirectory.deleteOnExit();
							parentDirectory.mkdirs();
						}
						tempFile = new File(parentPath, tempName);
						os = new FileOutputStream(tempFile);

						// Handle spaces in the remote resource path
						sb = new StringBuffer();
						sb.append(resourceUrl.substring(0, indexOfUrn + RESOURCE_URN.length()));
						sb.append(URLEncoder.encode(resourceUrl.substring(indexOfUrn + RESOURCE_URN.length()), "UTF-8")); //$NON-NLS-1$

						Map<String, Object> response = requester.downloadResource(sb.toString(), os, -1);
						String responseCode = response.get("rqm_responseCode").toString();
						if (Integer.decode(responseCode) != 200) {
							String downloadErr = "";
							success = false;
							Set<String> keySet = response.keySet();
							Logger.Log.error("Shared resource download failed from URL " + sb.toString() + " - HTTP response set follows:");
							downloadErr = " Shared resource download failed from URL " + resourceUrl + " - HTTP response set follows:";
							progress.logProgress(10, "Download Failed.", null);
							for (String s : keySet) {
								Logger.Log.error(s + ": " + response.get(s));
								downloadErr=downloadErr.concat(s + ": " + response.get(s));
							}
							((CommandLineAdapter) toolAdapter).setErrorMessage(downloadErr);
							throw new Exception(downloadErr);
						} else {
							progress.logProgress(10, "Downloading Successful.", resource);
						}
						command = tempFile.getAbsolutePath();
					} catch (Exception e) {
						success = false;
						StringBuffer message = new StringBuffer("Failed to create temporary file for the command: ");
						message.append(e.getMessage());
						Logger.Log.error(message.toString(), e);
						progress.logProgress(10, "Download Failed.", null);
						((CommandLineAdapter)toolAdapter).uploadError(message.toString(), logResult);
						return;
					} finally {
						if (os != null){
							try {
								os.flush();
								os.close();
							} catch (IOException e) {
							}
						}
						if (!success){
							((CommandLineLogResult)logResult).setStatus(ResultStateEnum.ERROR);
							((CommandLineLogResult)logResult).setStartTime(System.currentTimeMillis());
							((CommandLineLogResult)logResult).setEndTime(System.currentTimeMillis());
							requester.createResult(null, element, logResult);
							progress.LogProgress(100, 500);
						}
					}
					if ('/' == File.separatorChar) {
						// If we are on UNIX/Linux, we need to set the executable bit on
						// the file
						try {
							Runtime.getRuntime().exec("chmod +x " + command);
						} catch (Exception e) {
							success = false;
							if (Logger.Log.isDebugEnabled()) {
								Logger.Log.debug("Failed to change file attributes: " + e.getMessage(), e);
							}
							progress.logProgress(100, "Downloaded File does not have Execute permission.", null);
						}
					}
				}
				else {

					// managed by the adapter
					// Get the command name
					if (command == null) {
						command = fullpath;
					}
				}
				if(!executionMonitor){
				progress.LogProgress(15, 200);
				progress.LogProgress(20, 200);
				progress.LogProgress(50, 200);
				}
				
				((CommandLineLogResult)logResult).setStatus(ResultStateEnum.INPROGRESS);
				((CommandLineLogResult)logResult).setStartTime(System.currentTimeMillis());
				
				//run Selenium/JUnit test.  command contains the JUnit testcase to run, args contains any classpath entries needed
				if (isJUnitSelenium) {
					junitReportFile = File.createTempFile("QMAdapterJUnitReport", ".ser"); //$NON-NLS-1$ //$NON-NLS-2$
					// TODO Probably should allow user to set the screenshot dir via a propety in selenium.properties
					junitScreenshotDir = new File(junitReportFile.getParentFile(), "QMAdapterJUnitImgs_" + System.currentTimeMillis()); //$NON-NLS-1$
					
					/*
					 * Gather the java system properties specified in the testscript and any adapter level properties from selenium.properties.  
					 * We put any from selenium.properties at the end of the list so they will over-ride any with same name from testscript.
					 */
					StringBuffer javaProperties = new StringBuffer();
					String scriptProps = requester.getValue("qmtask:properties", element);
					String adapterProps = CommandLinePreferences.prefs().getProperty(CommandLinePreferences.JUNIT_SYSTEM_PROPERTIES, null);
					if (scriptProps != null)
						javaProperties.append(scriptProps);
					if (adapterProps != null) {
						if (javaProperties.length() > 0)
							javaProperties.append(",");
						javaProperties.append(adapterProps);
					}
					
					/*
					 * command is the junit class to run, args contains the classpath entries required to run it.
					 * seleniumClasspath contains the Selenium and JUnit classpath entries required.  
					 */
					StringBuffer javaArgs = new StringBuffer();
					
					// Java JVM debugging args for connecting debugger 
					// javaArgs.append("-Xdebug -Xnoagent -Djava.compiler=NONE -Xrunjdwp:transport=dt_socket,server=y,suspend=y,address=8000 ");
					
					String seleniumClasspath = CommandLinePreferences.prefs().getProperty(CommandLinePreferences.SELENIUM_CLASSPATH, null);
					String junitClasspath = CommandLinePreferences.prefs().getProperty(CommandLinePreferences.JUNIT_CLASSPATH, null);
					javaArgs.append("-cp \""); //$NON-NLS-1$
					javaArgs.append(cmdlineAdapterClasspathEntry());
					if (args != null) {
						javaArgs.append(File.pathSeparator);
						javaArgs.append(args.trim());
					}
					if (junitClasspath != null) {
						javaArgs.append(File.pathSeparator);
						javaArgs.append(junitClasspath.trim());
					}
					if (seleniumClasspath != null) {
						javaArgs.append(File.pathSeparator);
						javaArgs.append(seleniumClasspath.trim());
					}
					javaArgs.append("\" com.ibm.rqm.adapter.commandline.JUnitLauncher "); //$NON-NLS-1$
					javaArgs.append(JUnitLauncher.COMMAND_LINE_ARGUMENT_TEST);
					javaArgs.append(" \""); //$NON-NLS-1$
					javaArgs.append(command);
					javaArgs.append("\" "); //$NON-NLS-1$
					javaArgs.append(JUnitLauncher.COMMAND_LINE_ARGUMENT_REPORT_FILE);
					javaArgs.append(" \""); //$NON-NLS-1$
					javaArgs.append(junitReportFile.getCanonicalPath());
					javaArgs.append("\" "); //$NON-NLS-1$
					javaArgs.append(JUnitLauncher.COMMAND_LINE_ARGUMENT_SCREENSHOT_DIR);
					javaArgs.append(" \""); //$NON-NLS-1$
					javaArgs.append(junitScreenshotDir.getCanonicalPath());
					javaArgs.append("\""); //$NON-NLS-1$
					if (javaProperties.length() > 0) {
						javaArgs.append(" "); //$NON-NLS-1$
						javaArgs.append(JUnitLauncher.COMMAND_LINE_ARGUMENT_PROPERTIES);
						javaArgs.append(" \""); //$NON-NLS-1$
						javaArgs.append(javaProperties.toString());
						javaArgs.append("\""); //$NON-NLS-1$
					}
					if (Logger.Log.isDebugEnabled()) {
						Logger.Log.debug("Executing command: java " + javaArgs.toString());
					}
					toolAdapter.runTest(new String[] { "java", javaArgs.toString() }, requester.getExecutionVariables(element), true); //$NON-NLS-1$
				}
				// run generic command line test
				else {
					toolAdapter.runTest(new String[] { command, args }, requester.getExecutionVariables(element), false);
				}
				toolAdapter.setStatus(CommandLineAdapter.STATUS_EXECUTION_DONE);
				
				((CommandLineLogResult)logResult).setEndTime(System.currentTimeMillis());
				File stdOutFile = ((CommandLineAdapter)toolAdapter).getStdOutFile();
				File stdErrFile = ((CommandLineAdapter)toolAdapter).getStdErrFile();
				try {
					if (stdOutFile != null && stdOutFile.getCanonicalFile().length() !=0){
						logResult.addAttachment(stdOutFile.getCanonicalPath(), "text/plain"); //$NON-NLS-1$
					}
					if (stdErrFile != null && stdErrFile.getCanonicalFile().length()!=0){
						logResult.addAttachment(stdErrFile.getCanonicalPath(), "text/plain"); //$NON-NLS-1$
					}
				}
				catch(IOException e){
					Logger.Log.error(e.getMessage(), e);
				}

				if(!executionMonitor){
					progress.LogProgress(70, 200);
					progress.LogProgress(80, 200);
					progress.LogProgress(90, 200);
					progress.logProgress(99, null, null);
				}
			} catch (Exception e) {
				success = false;
				Logger.Log.error(e.getMessage(), e);
			}

			try {
				// Load execution variables which might have been updated by adapter execution
				Properties execVars = null;
				File execVarsFile = toolAdapter.getExecVariablesFile();
				if (execVarsFile != null) {
					InputStream varsIn = null;
					try {
						varsIn = new FileInputStream(execVarsFile);
						execVars = new Properties();
						execVars.load(varsIn);
						varsIn.close();
						execVarsFile.delete();
					} catch (FileNotFoundException fnfe) {
						execVars = null;
						Logger.Log.error("Failed to find execution variables file " + execVarsFile.getPath() + ".  Unable to pass on any modified variable values."); //$NON-NLS-1$ //$NON-NLS-2$
					} catch (IOException ioe) {
						execVars = null;
						Logger.Log.error("Errors encountered trying to load execution variables after execution of task.  Unable to pass on any modified variable values. " + ioe.getLocalizedMessage()); //$NON-NLS-1$
					}
				}
				// Create result and update adapter request
				if (isJUnitSelenium) {
					FileInputStream fis = null;
					ObjectInputStream ois = null;
					try {
						fis = new FileInputStream(junitReportFile);
						ois = new ObjectInputStream(fis);
						JUnitReport report = (JUnitReport)ois.readObject();
						JUnitReportLogger jLogger = new JUnitReportLogger(logResult, requester);
						int numFailures = jLogger.log(report);
						logResult.setStatus(numFailures > 0 ? ResultStateEnum.FAILED : ResultStateEnum.PASSED);
					} catch (FileNotFoundException fnfe) {
						StringBuffer sb = new StringBuffer();
						sb.append("Failed to find JUnit report file "); //$NON-NLS-1$
						sb.append(junitReportFile.getPath());
						sb.append(".  Result will not contain step results for individual JUnit tests."); //$NON-NLS-1$
						Logger.Log.error(sb.toString()); 
						logResult.setStatus(ResultStateEnum.ERROR);
					} catch (Exception e) {
						StringBuffer sb = new StringBuffer();
						sb.append("Failed to read serialized JUnit report object: ");
						sb.append(e.getLocalizedMessage());
						sb.append("  Execution result will not contain step results for individual JUnit tests.");
						Logger.Log.error(sb.toString());
						logResult.setStatus(ResultStateEnum.ERROR);
					} finally {
						try {
							if (ois != null)
								ois.close();
							if (fis != null)
								fis.close();
							if (junitReportFile.exists())
								junitReportFile.delete();
							/*
							 * Since we remove the screenshot files after we
							 * attach them to the result, we should be able to remove the
							 * (empty) screenshot directory, unless it
							 * contains other files we don't expect.
							 */
							if (junitScreenshotDir.exists() && junitScreenshotDir.isDirectory())
								junitScreenshotDir.delete();
						} catch (IOException ioe) {
							// Do nothing
						}
					}
					requester.createResultWithElements(null, element, logResult);
				}
				else
					requester.createResult(null, element, logResult);
				String message = (success ? "Done !" : "Done with error(s) !");
				progress.logProgress(100, message, null, execVars);
			} catch (AdapterException e) {
				Logger.Log.error("Failed to log an execution complete with the server: " + e.getMessage(), e);
			}

			deleteFile(tempFile);
			if((tempFile!= null && !tempFile.exists()) && (parentDirectory!= null && parentDirectory.exists())){
				deleteFile(parentDirectory);
			}
		} catch(ExecutionStoppedException e){
			Logger.Log.error("Excecution cancelled under UI: " + e.getMessage(), e);
		}
		finally {
			processor.setBusy(false); 
			((CommandLineAdapter)toolAdapter).cleanupOutFiles();
			logResult = null;
			toolAdapter = null;
		}
	  }
	}
	
	private static void deleteFile(File file) {
		if (file != null) {
			file.delete();
		}
	}
	
	public CommandLineAdapter getCommandLineAdapter()
	{
		return this.toolAdapter;
	}
	public String getTaskId()
	{
		return taskId;
	}
	
	/**
	 * Look in the current directory for the com.ibm.rqm.adapter.commandline jar file.
	 * For development environments we return "./bin".
	 * 
	 * @return Path to the jar file
	 */
	private String cmdlineAdapterClasspathEntry() {
		// Check if operating in a development environment
		if (System.getProperty("com.ibm.rqm.isDevEnv") != null) //$NON-NLS-1$
			return "./bin"; //$NON-NLS-1$
			
		File fileList[] = new File(".").listFiles(); //$NON-NLS-1$
		if (fileList != null) {
			for (File f : fileList) {
				if (f.getName().startsWith("com.ibm.rqm.adapter.commandline")) //$NON-NLS-1$
					return f.getName();
			}
		}
		Logger.Log.error("Failed to find com.ibm.rqm.adapter.commandline jar file");
		return null;
	}
}
