/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2012. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.rqm.adapter.commandline;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import com.ibm.rqm.adapter.library.data.ResultStateEnum;

/**
 * The user can control certain aspects of the command line execution through properties set in the
 * command line properties file.  These properties set things like the exit code from the application
 * under test or additional properties that should be included in the execution result.
 */
public class CommandLinePreferences {
	
	private File commandLinePropertiesFile = null;
	private Properties commandLineProperties = null ;
	private boolean propertiesLoaded = false;
	private Map<String,ResultStateEnum> exitCodes = null;
	private Map<String,ResultStateEnum> defaultExitCodes = null;

	/**
	 * The property used to decide if default system properties should be added to the execution
	 * result record generated by an execution.
	 */
	public static final String INCLUDE_SYSTEM_PROPERTIES = "com.ibm.rqm.commandline.systemproperties";
	
	/**
	 * The property used to decide if runtime updates to the execution result are supported by the
	 * application under test.  If set to a valid file specification then additional processing
	 * periodically monitors changes in state of this file and updates the execution result state
	 * on change.  This file must be in properties file format, with two properties RQM tracks
	 * completion.current and completion.total, that are used to formulate the total execution.
	 */
	public static final String EXECUTION_PROPERTIES = "com.ibm.rqm.commandline.executionproperties";
	public static final String ALLOWED_EXECUTABLES = "com.ibm.rqm.commandline.allowed_executables";
	
	public static final String SELENIUM_CLASSPATH = "com.ibm.rqm.selenium.seleniumClasspath";
	public static final String JUNIT_SYSTEM_PROPERTIES = "com.ibm.rqm.junit.system.properties";
	public static final String JUNIT_CLASSPATH = "com.ibm.rqm.junit.classpath";
	
	private static final String EXIT_CODE_PREFIX = "com.ibm.rqm.commandline.exitcode.";
	private static final String RQM_PREFIX = "com.ibm.rqm.";
	private static final String OPTIONNOTSET = "<CommandLineProperties Not Set>" ;
	
	private static CommandLinePreferences preferences = null;
	
	/**
	 * There should only ever be one set of preferences for a single adapter, so make it a 
	 * singleton that is accessible from here.
	 */
	protected static CommandLinePreferences prefs() {
		if ( preferences == null )
			new CommandLinePreferences(null); // initializes preferences indirectly
		return preferences;
	}
	
	/**
	 * The constructor for the property set 
	 * @param prefsFile
	 */
	public CommandLinePreferences(File prefsFile) {
		if ( prefsFile != null && prefsFile.exists() ) {
			commandLinePropertiesFile = prefsFile;
			// Load the properties from the ivory.properties file.
			try {
				FileInputStream propertyFile = new FileInputStream(commandLinePropertiesFile);
				if ( propertyFile != null ) {
					commandLineProperties = new java.util.Properties() ;
					commandLineProperties.load(propertyFile) ;
				}
			}
 			catch(IOException e) { // propertiesLoaded = false
			}			
		} else {
			commandLinePropertiesFile = null;
		}
		// there should only ever be one version of the preferences, so capture on construction ...
		preferences = this;
	}
	
	/**
	 * Check if the specified file name was able to be loaded into a property set or not.
	 * 
	 * @return true if the specified file name denotes a valid property file.
	 */
	public boolean getPropertiesLoaded() {
		return this.propertiesLoaded;
	}
	
	/**
	 * Return the value of an Command Line property.
	 *
	 * @param option    The property to retrieve
	 */
	public String getProperty(String property, String def) {
		String value = null ;
		if ( commandLineProperties != null ) {
			value = commandLineProperties.getProperty(property, OPTIONNOTSET) ;
			value = ( value != null && !value.equals(OPTIONNOTSET) ? value.trim() : def );
		}
		return value;
	}

	/**
	 * Return the value of an Command Line property.
	 *
	 * @param option    The property to retrieve
	 */
	public boolean getBooleanProperty(String property, boolean def) {
		String value = getProperty(property, null);
		return (value != null ? Boolean.parseBoolean(value) : def);
	}

	/**
	 * Return the value of an Command Line property.
	 *
	 * @param option    The property to retrieve
	 */
	public int getIntProperty(String property, int def) {
		String value = getProperty(property, null);
		return ( value != null ? Integer.parseInt(value) : def);
	}

	/**
	 * Get the value associated with a exit code as defined in the command line adapter
	 * properties file.  A <code>null</code> value is returned if there is no mapping for
	 * the specified exit code.
	 * 
	 * @param	exitCode	The exit code looked for in the property set.
	 * @return	The associated exit result or <code>null</code> if no mapping is available.
	 */
	public ResultStateEnum getExitCode(int exitCode) {
		ResultStateEnum result = null;
		Map<String,ResultStateEnum> codes = getExitCodes();
		Map<String,ResultStateEnum> defaultCodes = getDefaultExitCodes();
		if ( codes != null ) {
			String exitCodeString = Integer.toString(exitCode);
			ResultStateEnum value = codes.get(exitCodeString);
			if ( value == null ) {
				value = codes.get("*");
			}
			value = ( value != null ? value : defaultCodes.get(exitCodeString) );
			result = ( value != null ? value : ResultStateEnum.INCONCLUSIVE );
		} else { // use default settings 0 == PASS and * == FAIL
			result = ( exitCode == 0 ? ResultStateEnum.PASSED : ResultStateEnum.FAILED );
		}
		return result;
	}
	
	private Map<String,ResultStateEnum> getExitCodes() {
		if ( this.exitCodes == null && commandLineProperties != null ) {
			Enumeration<?> names = commandLineProperties.propertyNames();
			Map<String,ResultStateEnum> codes = new HashMap<String,ResultStateEnum>();
			boolean includesExitCodes = false;
			while ( names.hasMoreElements() ) {
				String name = (String)names.nextElement();
				if ( name != null && name.startsWith(EXIT_CODE_PREFIX) ) {
					String exitCodeName = name.substring(name.lastIndexOf('.')+1);
					String exitCode = commandLineProperties.getProperty(name).trim();
					codes.put(exitCode, findExitCode(exitCodeName.toUpperCase()));
					includesExitCodes = true;
				}
			}
			if ( includesExitCodes )
				this.exitCodes = codes;
		}
		return this.exitCodes;
	}
	
	private ResultStateEnum findExitCode(String exitCodeName) {
		Map<String,ResultStateEnum> defaultCodes = getDefaultExitCodes();
		ResultStateEnum result = defaultCodes.get(exitCodeName);
		return ( result != null ? result : ResultStateEnum.UNKNOWN );
	}
	
	private Map<String,ResultStateEnum> getDefaultExitCodes() {
		if ( this.defaultExitCodes == null ) {
			this.defaultExitCodes = new HashMap<String,ResultStateEnum>();
			this.defaultExitCodes.put("PASSED", ResultStateEnum.PASSED);
			this.defaultExitCodes.put("FAILED", ResultStateEnum.FAILED);
			this.defaultExitCodes.put("ERROR", ResultStateEnum.ERROR);
			//this.defaultExitCodes.put("INCOMPLETE", ResultStateEnum.INCOMPLETE);
			this.defaultExitCodes.put("INCONCLUSIVE", ResultStateEnum.INCONCLUSIVE);
			this.defaultExitCodes.put("BLOCKED", ResultStateEnum.BLOCKED);
			//this.defaultExitCodes.put("PARTIALLY_BLOCKED", ResultStateEnum.PARTIALLY_BLOCKED);
			//this.defaultExitCodes.put("PERMANENTLY_FAILED", ResultStateEnum.PERMANENTLY_FAILED);
			this.defaultExitCodes.put("INPROGRESS", ResultStateEnum.INPROGRESS);
		}
		return this.defaultExitCodes;
	}
	
	/**
	 * Returns a Map of non-RQM properties that should be included in the Execution
	 * Result record generated from this execution.  An empty map is returned if
	 * there are no non-RQM properties in the command line properties file.
	 * 
	 * @return	A map of non-RQM properties or an empty map if no non-RQM are present.
	 */
	public Map<String,String> getNonRqmProperties() {
		Map<String,String> result = new HashMap<String,String>();
		if ( commandLineProperties != null ) {
			Enumeration<?> names = commandLineProperties.propertyNames();
			while ( names.hasMoreElements() ) {
				String name = (String)names.nextElement();
				if ( name != null && !name.startsWith(RQM_PREFIX) ) {
					result.put(name, (String)commandLineProperties.get(name));
				}
			}
		}
		return result;
	}

}
