/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2011, 2012. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.rqm.adapter.commandline;

import java.io.File;
import java.io.IOException;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import com.ibm.rqm.adapter.library.data.AbstractLogEvent;
import com.ibm.rqm.adapter.library.data.AbstractLogResult;
import com.ibm.rqm.adapter.library.data.AdapterConstants;
import com.ibm.rqm.adapter.library.data.CustomProperty;
import com.ibm.rqm.adapter.library.data.ElementTypeEnum;
import com.ibm.rqm.adapter.library.data.ILogEvent;
import com.ibm.rqm.adapter.library.data.Logger;
import com.ibm.rqm.adapter.library.data.Messages;
import com.ibm.rqm.adapter.library.data.NewRequester;
import com.ibm.rqm.adapter.library.data.ResultStateEnum;

/**
 * 
 * @author tneal
 *
 * Read the JUnitReport object and create logEvents from each JUnitReportEntry found.
 */
public class JUnitReportLogger {

	public AbstractLogResult log = null;
	public NewRequester requester = null;

	public JUnitReportLogger(AbstractLogResult log, NewRequester requester) {
		this.log = log;
		this.requester = requester;
	}

	public AbstractLogResult getLog() {
		return log;
	}

	public void setLog(AbstractLogResult log) {
		this.log = log;
	}

	/**
	 * Add a logEntry to the log for each JUnit test in report
	 * 
	 * @param report
	 *            JUnitReport from execution of JUnit testcases/testsuites
	 * 
	 * @return number of failures
	 */
	public int log(JUnitReport report) {
		int entriesLogged = 0;
		int testsFailed = 0;
		int testsPassed = 0;
		if (log == null || report == null) {
			return 0;
		}
		List<AbstractLogEvent> logEvents = new ArrayList<AbstractLogEvent>();
		for (JUnitTestCaseReporter tcReporter : report.getTestCaseReporters()) {
			for (JUnitReportEntry reportEntry : tcReporter.getReportEntries()) {
				entriesLogged++;
				AbstractLogEvent event = new AbstractLogEvent();
				event.starttime = reportEntry.getStartTimestamp();
				event.endtime = reportEntry.getEndTimestamp();
				event.resultState = reportEntry.getType() == JUnitReportEntryType.PASS ? ResultStateEnum.PASSED	: ResultStateEnum.FAILED;
				if (event.resultState == ResultStateEnum.PASSED)
					testsPassed++;
				else
					testsFailed++;
				event.type = ElementTypeEnum.REPORTING;
				event.description = tcReporter.getTestCaseName() + ":" + reportEntry.getDescription(); //$NON-NLS-1$
				if (reportEntry.getExceptionTraceback() != null) {
					StringBuffer sb = new StringBuffer();
					sb.append(event.description);
					sb.append(System.getProperty("line.separator")); //$NON-NLS-1$
					sb.append(reportEntry.getExceptionTraceback());
					event.description = sb.toString();
				}
				// Handle screenshots
				int screenShotCnt = 1;
				ArrayList<CustomProperty> custPropList = new ArrayList<CustomProperty>(Arrays.asList(log.getCustomProperties()));
				for (File screenshot : reportEntry.getScreenshots()) {
					// Attach screenshot to result.  Once it's attached, remove it.
					String path = screenshot.getAbsolutePath();
					String aId = null;
					try {
						/* 
						 * Prefix the customField/label of the screenshot with "RQMHIDDEN_" so it's not displayed in 
						 * the Attachments section of the execution result.  It will be displayed with the associated
						 * JUnit test in the step result.
						 */
						aId = requester.createAttachment(path, URLConnection.guessContentTypeFromName(path), URLEncoder.encode("RQMHIDDEN_" + screenshot.getName(), "UTF-8")); //$NON-NLS-1$ //$NON-NLS-2$
						log.addAttachmentId(aId);
						screenshot.delete();
					} catch (IOException e) {
						Logger.Log.error(Messages.getString("NewRequester.296") + path + e.getLocalizedMessage()); //$NON-NLS-1$
						continue;
					}
					// Create CustomProperty with name StepX_ScreenshotY 
					// and value is UUID of associated attachment
					CustomProperty custProp = new CustomProperty();
					StringBuffer propName = new StringBuffer();
					propName.append("Step"); //$NON-NLS-1$
					propName.append(String.valueOf(entriesLogged));
					propName.append("_"); //$NON-NLS-1$
					propName.append("Screenshot"); //$NON-NLS-1$
					propName.append(String.valueOf(screenShotCnt++));
					custProp.setPropertyName(propName.toString());
					custProp.setPropertyValue(aId);
					custProp.setNameSpaceWithPrefix("rqmscreenshot", AdapterConstants.RESULT_SCREENSHOT_NAMESPACE); //$NON-NLS-1$
					custPropList.add(custProp);
				}
				if (screenShotCnt > 1)
					log.setCustomProperties(custPropList.toArray(new CustomProperty[custPropList.size()]));
				logEvents.add(event);
			}
		}
		this.log.setLogEvents(logEvents.toArray(new ILogEvent[] {}));
		return testsFailed;
	}
}
