/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2011. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.rqm.adapter.commandline;

import java.io.File;
import java.util.ArrayList;

import org.apache.abdera.model.Element;

import com.ibm.rqm.adapter.library.data.AdapterException;
import com.ibm.rqm.adapter.library.data.Logger;
import com.ibm.rqm.adapter.library.data.NewRequester;
import com.ibm.rqm.adapter.library.data.RecordedScript;
import com.ibm.rqm.adapter.library.data.RecordedStep;
import com.ibm.rqm.adapter.library.data.RichTextAttachmentInfo;
import com.ibm.rqm.adapter.library.data.RecordedStep.StepType;
import com.ibm.rqm.adapter.library.util.AdapterUtil;

/**
 * Sample implementation of a recorder.  This is a very basic implementation intended to give guidance on the APIs in the adapter library
 * that would be used by a recorded.  It is not meant to serve as a full guide to developing a recorder for RQM.  As it is missing several
 * important features (i.e. actual recording, support for cancel operation, etc..).
 * 
 * To run this:
 * 	1) Uncomment QMCommandLineProcessor.processRecord.
 *	2) Add RECORD_CAPABILITY to QMCommandLineAdapter.ADAPTER_CAPABILITIES
 */
public class SampleRecorder {
	
	public void processRecord(NewRequester requester, String url, Element task){	
		String scriptUrl = null;
		RecordedScript script = new RecordedScript();
		ArrayList<RecordedStep> liveSteps = new ArrayList<RecordedStep>();
		
		try{
			script.setDescription("recorded.sample.script");
			script.getSteps().add(recordSimpleStep());
			script.getSteps().add(recordTranslatedUnescapedStep());
		
    		scriptUrl = requester.uploadRecordedScript(task, script, null);

    		// Now append addition step(s) during live recording
    		liveSteps.add(recordImageStep(requester));
    		requester.appendRecordedSteps(task, liveSteps, scriptUrl);
    		liveSteps.clear(); // always clear after calling appendRecordedSteps
    		
    		// Record more step(s)
    		liveSteps.add(recordAttachmentStep());
    		requester.appendRecordedSteps(task, liveSteps, scriptUrl);
    		liveSteps.clear();
		} catch (AdapterException e) {
			Logger.Log.error(e.getMessage(), e);
		} finally {
			try{
				// Finish with recording, so post progress as 100%.  Always do in finally to always tell RQM the
				// recording is done.  Even in presence of error condition.
				requester.logProgress(url, 100, 200, "Sample recording completed successfully.", null, null);
			} catch (AdapterException e) {
				Logger.Log.error(e.getMessage(), e);
			}
		}
	}
	
	private RecordedStep recordSimpleStep(){
        RecordedStep step = new RecordedStep();
        
        step.setDescription("Click on users home page.");
        step.setExpectedResult("home page is displayed");

        return step;
	}
	
	private RecordedStep recordTranslatedUnescapedStep(){
        RecordedStep step = new RecordedStep();

        step.setDescription("Description with illegal xml characters <>&; should be fine.");

        // Use Unicode notation to get characters into the string.  The chars in the string have no special escaping\encoding, except Java's native UTF16 encoding.
        step.setExpectedResult("\u30a2\u30c0\u30d7\u30bf\u30fc\u306f\u63a5\u7d9a\u3055\u308c\u307e\u3057\u305f\u3002");
        
        return step;
	}
	
	private RecordedStep recordImageStep(NewRequester requester){
		
		RecordedStep step = new RecordedStep();
	
		File dImage = new File("samples/TestImage1.jpg");
		File eImage = new File("samples/TestImage2.jpg");

		if(!dImage.exists() || !eImage.exists()){
			step.setDescription("Unable to find sample images.");
			return step;
		}
		
		// Quick way to set step field to contain only an image.
		RichTextAttachmentInfo dImageLinkInfo = requester.createAttachmentForRichTextHref(dImage.getAbsolutePath(), "image/jpeg");		
		step.setDescriptionRichText(AdapterUtil.getImageXhtml(dImage.getName(), dImageLinkInfo.getAttachmentId(), dImageLinkInfo.getAttachmentUrl()));
		
		// Manually construct XHTML, element by element.  This approach does not require any escaping of strings and will always produce valid XML.
		RichTextAttachmentInfo eImageLinkInfo = requester.createAttachmentForRichTextHref(eImage.getAbsolutePath(), "image/jpeg");	
		
		Element div = AdapterUtil.newXmlElement("div", null);
        div.setText("No need to escape text set on XML elements<>&;.  There should be an image after this text: ");
        Element a = AdapterUtil.newXmlElement("a", div);
        a.setAttributeValue("href", eImageLinkInfo.getAttachmentUrl());
        Element img = AdapterUtil.newXmlElement("img", a);
        img.setAttributeValue("border", "0");
        img.setAttributeValue("id", eImageLinkInfo.getAttachmentId());
        img.setAttributeValue("alt", eImage.getName());
        img.setAttributeValue("src", eImageLinkInfo.getAttachmentUrl());
        step.setExpectedResultRichText(div);
        
        return step;
	}
	
	private RecordedStep recordAttachmentStep(){
		
		RecordedStep step = new RecordedStep();
	
		File file 	= new File("samples/TestAttachment.txt");
		String contentType = "text/plain";

		if(!file.exists()){
			step.setDescription("Unable to find sample attachment.");
			return step;
		}
		
		step.setDescription("Step with attachment");
		step.setExpectedResult("Validate this step has an attachment");
        step.setType(StepType.reporting);
		
        step.addAttachment(file.getAbsolutePath(), contentType);
        
        return step;
	}
}